@@ -7,6 +7,9 @@
 /* Version number. */
 #undef VERSION
 
+/* Hack for Visual C++ suggested by irox. */
+#undef alloca
+
 /*
  * Internationalization stuffs.
  */
@@ -20,3 +23,8 @@
 #undef HAVE_GETTEXT
 
 #undef HAVE_LC_MESSAGES
+
+/*
+ * DOS specific
+ */
+#undef HAVE_DOS_FILE_NAMES
