@@ -1,5 +1,5 @@
 /* grep.c - main driver file for grep.
-   Copyright (C) 1992, 1997, 1998 Free Software Foundation, Inc.
+   Copyright (C) 1992, 1997, 1998, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -35,10 +35,20 @@
 #include "getopt.h"
 #include "getpagesize.h"
 #include "grep.h"
+#include "savedir.h"
 
 #undef MAX
 #define MAX(A,B) ((A) > (B) ? (A) : (B))
 
+struct stats
+{
+  struct stats *parent;
+  struct stat stat;
+};
+
+/* base of chain of stat buffers, used to detect directory loops */
+static struct stats stats_base;
+
 /* if non-zero, display usage information and exit */
 static int show_help;
 
@@ -52,8 +62,9 @@ static struct option long_options[] =
   {"basic-regexp", no_argument, NULL, 'G'},
   {"before-context", required_argument, NULL, 'B'},
   {"byte-offset", no_argument, NULL, 'b'},
-  {"context", no_argument, NULL, 'C'},
+  {"context", optional_argument, NULL, 'C'},
   {"count", no_argument, NULL, 'c'},
+  {"directories", required_argument, NULL, 'd'},
   {"extended-regexp", no_argument, NULL, 'E'},
   {"file", required_argument, NULL, 'f'},
   {"files-with-matches", no_argument, NULL, 'l'},
@@ -67,9 +78,11 @@ static struct option long_options[] =
   {"no-filename", no_argument, NULL, 'h'},
   {"no-messages", no_argument, NULL, 's'},
   {"quiet", no_argument, NULL, 'q'},
+  {"recursive", no_argument, NULL, 'r'},
   {"regexp", required_argument, NULL, 'e'},
   {"revert-match", no_argument, NULL, 'v'},
   {"silent", no_argument, NULL, 'q'},
+  {"text", no_argument, NULL, 'a'},
 #if O_BINARY
   {"binary", no_argument, NULL, 'U'},
   {"unix-byte-offsets", no_argument, NULL, 'u'},
@@ -81,86 +94,100 @@ static struct option long_options[] =
 };
 
 /* Define flags declared in grep.h. */
-char *matcher;
+char const *matcher;
 int match_icase;
 int match_words;
 int match_lines;
 
 /* For error messages. */
 static char *prog;
-static char *filename;
+static char const *filename;
 static int errseen;
 
-static int ck_atoi PARAMS((char const *, int *));
-static void usage PARAMS((int));
-static void error PARAMS((const char *, int));
-static int  setmatcher PARAMS((char *));
-static void reset PARAMS((int));
-static int  fillbuf PARAMS((size_t));
-static int  grepbuf PARAMS((char *, char *));
-static void prtext PARAMS((char *, char *, int *));
-static void prpending PARAMS((char *));
-static void prline PARAMS((char *, char *, int));
-static void nlscan PARAMS((char *));
-static int  grep PARAMS((int));
+/* How to handle directories.  */
+static enum
+  {
+    READ_DIRECTORIES,
+    RECURSE_DIRECTORIES,
+    SKIP_DIRECTORIES
+  } directories;
+
+static int  ck_atoi PARAMS ((char const *, int *));
+static void usage PARAMS ((int)) __attribute__((noreturn));
+static void error PARAMS ((const char *, int));
+static int  setmatcher PARAMS ((char const *));
+static char *page_alloc PARAMS ((size_t, char **));
+static int  reset PARAMS ((int, char const *, struct stats *));
+static int  fillbuf PARAMS ((size_t, struct stats *));
+static int  grepbuf PARAMS ((char *, char *));
+static void prtext PARAMS ((char *, char *, int *));
+static void prpending PARAMS ((char *));
+static void prline PARAMS ((char *, char *, int));
+static void print_offset_sep PARAMS ((off_t, int));
+static void nlscan PARAMS ((char *));
+static int  grep PARAMS ((int, char const *, struct stats *));
+static int  grepdir PARAMS ((char const *, struct stats *));
+static int  grepfile PARAMS ((char const *, struct stats *));
+#if O_BINARY
+static inline int undossify_input PARAMS ((register char *, size_t));
+#endif
 
 /* Functions we'll use to search. */
-static void (*compile) PARAMS((char *, size_t));
-static char *(*execute) PARAMS((char *, size_t, char **));
+static void (*compile) PARAMS ((char *, size_t));
+static char *(*execute) PARAMS ((char *, size_t, char **));
 
 /* Print a message and possibly an error string.  Remember
    that something awful happened. */
 static void
-error(mesg, errnum)
+error (mesg, errnum)
      const char *mesg;
      int errnum;
 {
   if (errnum)
-    fprintf(stderr, "%s: %s: %s\n", prog, mesg, strerror(errnum));
+    fprintf (stderr, "%s: %s: %s\n", prog, mesg, strerror (errnum));
   else
-    fprintf(stderr, "%s: %s\n", prog, mesg);
+    fprintf (stderr, "%s: %s\n", prog, mesg);
   errseen = 1;
 }
 
-/* Like error(), but die horribly after printing. */
+/* Like error (), but die horribly after printing. */
 void
-fatal(mesg, errnum)
+fatal (mesg, errnum)
      const char *mesg;
      int errnum;
 {
-  error(mesg, errnum);
-  exit(2);
+  error (mesg, errnum);
+  exit (2);
 }
 
 /* Interface to handle errors and fix library lossage. */
 char *
-xmalloc(size)
+xmalloc (size)
      size_t size;
 {
   char *result;
 
-  result = malloc(size);
+  result = malloc (size);
   if (size && !result)
-    fatal(_("memory exhausted"), 0);
+    fatal (_("memory exhausted"), 0);
   return result;
 }
 
 /* Interface to handle errors and fix some library lossage. */
 char *
-xrealloc(ptr, size)
+xrealloc (ptr, size)
      char *ptr;
      size_t size;
 {
   char *result;
 
   if (ptr)
-    result = realloc(ptr, size);
+    result = realloc (ptr, size);
   else
-    result = malloc(size);
+    result = malloc (size);
   if (size && !result)
-    fatal(_("memory exhausted"), 0);
+    fatal (_("memory exhausted"), 0);
   return result;
-
 }
 
 /* Convert STR to a positive integer, storing the result in *OUT.
@@ -184,55 +211,113 @@ ck_atoi (str, out)
    all reads aligned on a page boundary and multiples of the
    page size. */
 
+static char *ubuffer;		/* Unaligned base of buffer. */
 static char *buffer;		/* Base of buffer. */
 static size_t bufsalloc;	/* Allocated size of buffer save region. */
 static size_t bufalloc;		/* Total buffer size. */
 static int bufdesc;		/* File descriptor. */
 static char *bufbeg;		/* Beginning of user-visible stuff. */
 static char *buflim;		/* Limit of user-visible stuff. */
+static size_t pagesize;		/* alignment of memory pages */
 
 #if defined(HAVE_MMAP)
 static int bufmapped;		/* True for ordinary files. */
-static struct stat bufstat;	/* From fstat(). */
 static off_t bufoffset;		/* What read() normally remembers. */
+static off_t initial_bufoffset;	/* Initial value of bufoffset. */
 #endif
 
-/* Reset the buffer for a new file.  Initialize
-   on the first time through. */
-static void
-reset(fd)
-     int fd;
+/* Return VAL aligned to the next multiple of ALIGNMENT.  VAL can be
+   an integer or a pointer.  Both args must be free of side effects.  */
+#define ALIGN_TO(val, alignment) \
+  ((size_t) (val) % (alignment) == 0 \
+   ? (val) \
+   : (val) + ((alignment) - (size_t) (val) % (alignment)))
+
+/* Return the address of a new page-aligned buffer of size SIZE.  Set
+   *UP to the newly allocated (but possibly unaligned) buffer used to
+   *build the aligned buffer.  To free the buffer, free (*UP).  */
+static char *
+page_alloc (size, up)
+     size_t size;
+     char **up;
 {
-  static int initialized;
-
-  if (!initialized)
-    {
-      initialized = 1;
-#ifndef BUFSALLOC
-      bufsalloc = MAX(8192, getpagesize());
+  /* HAVE_WORKING_VALLOC means that valloc is properly declared, and
+     you can free the result of valloc.  This symbol is not (yet)
+     autoconfigured.  It can be useful to define HAVE_WORKING_VALLOC
+     while debugging, since some debugging memory allocators might
+     catch more bugs if this symbol is enabled.  */
+#if HAVE_WORKING_VALLOC
+  *up = valloc (size);
+  return *up;
 #else
-      bufsalloc = BUFSALLOC;
+  size_t asize = size + pagesize - 1;
+  if (size <= asize)
+    {
+      *up = malloc (asize);
+      if (*up)
+	return ALIGN_TO (*up, pagesize);
+    }
+  return NULL;
 #endif
+}
+
+/* Reset the buffer for a new file, returning zero if we should skip it.
+   Initialize on the first time through. */
+static int
+reset (fd, file, stats)
+     int fd;
+     char const *file;
+     struct stats *stats;
+{
+  if (pagesize == 0)
+    {
+      size_t ubufsalloc;
+      pagesize = getpagesize ();
+      if (pagesize == 0)
+	abort ();
+#ifndef BUFSALLOC
+      ubufsalloc = MAX (8192, pagesize);
+#else
+      ubufsalloc = BUFSALLOC;
+#endif
+      bufsalloc = ALIGN_TO (ubufsalloc, pagesize);
       bufalloc = 5 * bufsalloc;
       /* The 1 byte of overflow is a kludge for dfaexec(), which
 	 inserts a sentinel newline at the end of the buffer
 	 being searched.  There's gotta be a better way... */
-      buffer = valloc(bufalloc + 1);
-      if (!buffer)
-	fatal(_("memory exhausted"), 0);
+      if (bufsalloc < ubufsalloc
+	  || bufalloc / 5 != bufsalloc || bufalloc + 1 < bufalloc
+	  || ! (buffer = page_alloc (bufalloc + 1, &ubuffer)))
+	fatal (_("memory exhausted"), 0);
       bufbeg = buffer;
       buflim = buffer;
     }
   bufdesc = fd;
+
+  if (
 #if defined(HAVE_MMAP)
-  if (fstat(fd, &bufstat) < 0 || !S_ISREG(bufstat.st_mode))
+      1
+#else
+      directories != READ_DIRECTORIES
+#endif
+      )
+    if (fstat (fd, &stats->stat) != 0)
+      {
+	error ("fstat", errno);
+	return 0;
+      }
+  if (directories == SKIP_DIRECTORIES && S_ISDIR (stats->stat.st_mode))
+    return 0;
+#if defined(HAVE_MMAP)
+  if (!S_ISREG (stats->stat.st_mode))
     bufmapped = 0;
   else
     {
       bufmapped = 1;
-      bufoffset = lseek(fd, 0, 1);
+      bufoffset = initial_bufoffset = file ? 0 : lseek (fd, 0, 1);
     }
 #endif
+  return 1;
 }
 
 /* Read new stuff into the buffer, saving the specified
@@ -240,58 +325,55 @@ reset(fd)
    to the beginning of the buffer contents, and 'buflim'
    points just after the end.  Return count of new stuff. */
 static int
-fillbuf(save)
+fillbuf (save, stats)
      size_t save;
+     struct stats *stats;
 {
-  char *nbuffer, *dp, *sp;
   int cc;
 #if defined(HAVE_MMAP)
   caddr_t maddr;
 #endif
-  static int pagesize;
-
-  if (pagesize == 0 && (pagesize = getpagesize()) == 0)
-    abort();
 
   if (save > bufsalloc)
     {
+      char *nubuffer;
+      char *nbuffer;
+
       while (save > bufsalloc)
 	bufsalloc *= 2;
       bufalloc = 5 * bufsalloc;
-      nbuffer = valloc(bufalloc + 1);
-      if (!nbuffer)
-	fatal(_("memory exhausted"), 0);
+      if (bufalloc / 5 != bufsalloc || bufalloc + 1 < bufalloc
+	  || ! (nbuffer = page_alloc (bufalloc + 1, &nubuffer)))
+	fatal (_("memory exhausted"), 0);
+
+      bufbeg = nbuffer + bufsalloc - save;
+      memcpy (bufbeg, buflim - save, save);
+      free (ubuffer);
+      ubuffer = nubuffer;
+      buffer = nbuffer;
     }
   else
-    nbuffer = buffer;
-
-  sp = buflim - save;
-  dp = nbuffer + bufsalloc - save;
-  bufbeg = dp;
-  while (save--)
-    *dp++ = *sp++;
-
-  /* We may have allocated a new, larger buffer.  Since
-     there is no portable vfree(), we just have to forget
-     about the old one.  Sorry. */
-  buffer = nbuffer;
+    {
+      bufbeg = buffer + bufsalloc - save;
+      memcpy (bufbeg, buflim - save, save);
+    }
 
 #if defined(HAVE_MMAP)
   if (bufmapped && bufoffset % pagesize == 0
-      && bufstat.st_size - bufoffset >= bufalloc - bufsalloc)
+      && stats->stat.st_size - bufoffset >= bufalloc - bufsalloc)
     {
       maddr = buffer + bufsalloc;
-      maddr = mmap(maddr, bufalloc - bufsalloc, PROT_READ | PROT_WRITE,
+      maddr = mmap (maddr, bufalloc - bufsalloc, PROT_READ | PROT_WRITE,
 		   MAP_PRIVATE | MAP_FIXED, bufdesc, bufoffset);
       if (maddr == (caddr_t) -1)
 	{
           /* This used to issue a warning, but on some hosts
              (e.g. Solaris 2.5) mmap can fail merely because some
-             other process has an advisory erad lock on the file.
-             There's no point alarming the use about this misfeature */
+             other process has an advisory read lock on the file.
+             There's no point alarming the user about this misfeature.  */
 #if 0
-	  fprintf(stderr, _("%s: warning: %s: %s\n"), prog, filename,
-		  strerror(errno));
+	  fprintf (stderr, _("%s: warning: %s: %s\n"), prog, filename,
+		  strerror (errno));
 #endif
 	  goto tryread;
 	}
@@ -299,7 +381,7 @@ fillbuf(save)
       /* You might thing this (or MADV_WILLNEED) would help,
 	 but it doesn't, at least not on a Sun running 4.1.
 	 In fact, it actually slows us down about 30%! */
-      madvise(maddr, bufalloc - bufsalloc, MADV_SEQUENTIAL);
+      madvise (maddr, bufalloc - bufsalloc, MADV_SEQUENTIAL);
 #endif
       cc = bufalloc - bufsalloc;
       bufoffset += cc;
@@ -313,16 +395,17 @@ fillbuf(save)
       if (bufmapped)
 	{
 	  bufmapped = 0;
-	  lseek(bufdesc, bufoffset, 0);
+	  if (bufoffset != initial_bufoffset)
+	    lseek (bufdesc, bufoffset, 0);
 	}
-      cc = read(bufdesc, buffer + bufsalloc, bufalloc - bufsalloc);
+      cc = read (bufdesc, buffer + bufsalloc, bufalloc - bufsalloc);
     }
 #else
-  cc = read(bufdesc, buffer + bufsalloc, bufalloc - bufsalloc);
-#endif
+  cc = read (bufdesc, buffer + bufsalloc, bufalloc - bufsalloc);
+#endif /*HAVE_MMAP*/
 #if O_BINARY
-  if (O_BINARY && cc > 0)
-    cc = undossify_input(buffer + bufsalloc, cc);
+  if (cc > 0)
+    cc = undossify_input (buffer + bufsalloc, cc);
 #endif
   if (cc > 0)
     buflim = buffer + bufsalloc + cc;
@@ -332,6 +415,7 @@ fillbuf(save)
 }
 
 /* Flags controlling the style of output. */
+static int always_text;		/* Assume the input is always text. */
 static int out_quiet;		/* Suppress all normal output. */
 static int out_invert;		/* Print nonmatching stuff. */
 static int out_file;		/* Print filenames. */
@@ -339,14 +423,18 @@ static int out_line;		/* Print line numbers. */
 static int out_byte;		/* Print byte offsets. */
 static int out_before;		/* Lines of leading context. */
 static int out_after;		/* Lines of trailing context. */
+static int count_matches;	/* Count matching lines.  */
+static int list_files;		/* List matching files.  */
+static int no_filenames;	/* Suppress file names.  */
+static int suppress_errors;	/* Suppress diagnostics.  */
 
 /* Internal variables to keep track of byte count, context, etc. */
-static size_t totalcc;		/* Total character count before bufbeg. */
+static off_t totalcc;		/* Total character count before bufbeg. */
 static char *lastnl;		/* Pointer after last newline counted. */
 static char *lastout;		/* Pointer after last character output;
 				   NULL if no character has been output
 				   or if it's conceptually before bufbeg. */
-static size_t totalnl;		/* Total newline count before lastnl. */
+static off_t totalnl;		/* Total newline count before lastnl. */
 static int pending;		/* Pending lines of output. */
 static int done_on_match;		/* Stop scanning file on first match */
 
@@ -355,7 +443,7 @@ static int done_on_match;		/* Stop scanning file on first match */
 #endif
 
 static void
-nlscan(lim)
+nlscan (lim)
      char *lim;
 {
   char *beg;
@@ -367,35 +455,55 @@ nlscan(lim)
 }
 
 static void
-prline(beg, lim, sep)
+print_offset_sep (pos, sep)
+     off_t pos;
+     int sep;
+{
+  /* Do not rely on printf to print pos, since off_t may be longer than long,
+     and long long is not portable.  */
+
+  char buf[sizeof pos * CHAR_BIT];
+  char *p = buf + sizeof buf - 1;
+  *p = sep;
+
+  do
+    *--p = '0' + pos % 10;
+  while ((pos /= 10) != 0);
+
+  fwrite (p, 1, buf + sizeof buf - p, stdout);
+}
+
+static void
+prline (beg, lim, sep)
      char *beg;
      char *lim;
      int sep;
 {
   if (out_file)
-    printf("%s%c", filename, sep);
+    printf ("%s%c", filename, sep);
   if (out_line)
     {
-      nlscan(beg);
-      printf("%u%c", (unsigned int)++totalnl, sep);
+      nlscan (beg);
+      print_offset_sep (++totalnl, sep);
       lastnl = lim;
     }
   if (out_byte)
+    {
+      off_t pos = totalcc + (beg - bufbeg);
 #if O_BINARY
-    printf("%lu%c",
-	   (unsigned long int) dossified_pos(totalcc + (beg - bufbeg)), sep);
-#else
-    printf("%lu%c", (unsigned long int) (totalcc + (beg - bufbeg)), sep);
+      pos = dossified_pos (pos);
 #endif
-  fwrite(beg, 1, lim - beg, stdout);
-  if (ferror(stdout))
-    error(_("writing output"), errno);
+      print_offset_sep (pos, sep);
+    }
+  fwrite (beg, 1, lim - beg, stdout);
+  if (ferror (stdout))
+    error (_("writing output"), errno);
   lastout = lim;
 }
 
 /* Print pending lines of trailing context prior to LIM. */
 static void
-prpending(lim)
+prpending (lim)
      char *lim;
 {
   char *nl;
@@ -405,18 +513,18 @@ prpending(lim)
   while (pending > 0 && lastout < lim)
     {
       --pending;
-      if ((nl = memchr(lastout, '\n', lim - lastout)) != 0)
+      if ((nl = memchr (lastout, '\n', lim - lastout)) != 0)
 	++nl;
       else
 	nl = lim;
-      prline(lastout, nl, '-');
+      prline (lastout, nl, '-');
     }
 }
 
 /* Print the lines between BEG and LIM.  Deal with context crap.
    If NLINESP is non-null, store a count of lines between BEG and LIM. */
 static void
-prtext(beg, lim, nlinesp)
+prtext (beg, lim, nlinesp)
      char *beg;
      char *lim;
      int *nlinesp;
@@ -426,7 +534,7 @@ prtext(beg, lim, nlinesp)
   int i, n;
 
   if (!out_quiet && pending > 0)
-    prpending(beg);
+    prpending (beg);
 
   p = beg;
 
@@ -444,12 +552,12 @@ prtext(beg, lim, nlinesp)
       /* We only print the "--" separator if our output is
 	 discontiguous from the last output in the file. */
       if ((out_before || out_after) && used && p != lastout)
-	puts("--");
+	puts ("--");
 
       while (p < beg)
 	{
-	  nl = memchr(p, '\n', beg - p);
-	  prline(p, nl + 1, '-');
+	  nl = memchr (p, '\n', beg - p);
+	  prline (p, nl + 1, '-');
 	  p = nl + 1;
 	}
     }
@@ -459,19 +567,19 @@ prtext(beg, lim, nlinesp)
       /* Caller wants a line count. */
       for (n = 0; p < lim; ++n)
 	{
-	  if ((nl = memchr(p, '\n', lim - p)) != 0)
+	  if ((nl = memchr (p, '\n', lim - p)) != 0)
 	    ++nl;
 	  else
 	    nl = lim;
 	  if (!out_quiet)
-	    prline(p, nl, ':');
+	    prline (p, nl, ':');
 	  p = nl;
 	}
       *nlinesp = n;
     }
   else
     if (!out_quiet)
-      prline(beg, lim, ':');
+      prline (beg, lim, ':');
 
   pending = out_after;
   used = 1;
@@ -481,7 +589,7 @@ prtext(beg, lim, nlinesp)
    between matching lines if OUT_INVERT is true).  Return a count of
    lines printed. */
 static int
-grepbuf(beg, lim)
+grepbuf (beg, lim)
      char *beg;
      char *lim;
 {
@@ -498,36 +606,52 @@ grepbuf(beg, lim)
 	break;
       if (!out_invert)
 	{
-	  prtext(b, endp, (int *) 0);
+	  prtext (b, endp, (int *) 0);
 	  nlines += 1;
 	  if (done_on_match)
 	    return nlines;
 	}
       else if (p < b)
 	{
-	  prtext(p, b, &n);
+	  prtext (p, b, &n);
 	  nlines += n;
 	}
       p = endp;
     }
   if (out_invert && p < lim)
     {
-      prtext(p, lim, &n);
+      prtext (p, lim, &n);
       nlines += n;
     }
   return nlines;
 }
 
-/* Search a given file.  Return a count of lines printed. */
+/* Search a given file.  Normally, return a count of lines printed;
+   but if the file is a directory and we search it recursively, then
+   return -2 if there was a match, and -1 otherwise.  */
 static int
-grep(fd)
+grep (fd, file, stats)
      int fd;
+     char const *file;
+     struct stats *stats;
 {
   int nlines, i;
+  int not_text;
   size_t residue, save;
   char *beg, *lim;
 
-  reset(fd);
+  if (!reset (fd, file, stats))
+    return 0;
+
+  if (file && directories == RECURSE_DIRECTORIES
+      && S_ISDIR (stats->stat.st_mode))
+    {
+      /* Close fd now, so that we don't open a lot of file descriptors
+	 when we recurse deeply.  */
+      if (close (fd) != 0)
+	error (file, errno);
+      return grepdir (file, stats) - 2;
+    }
 
   totalcc = 0;
   lastout = 0;
@@ -538,13 +662,20 @@ grep(fd)
   residue = 0;
   save = 0;
 
+  if (fillbuf (save, stats) < 0)
+    {
+      if (! (is_EISDIR (errno, file) && suppress_errors))
+	error (filename, errno);
+      return nlines;
+    }
+
+  not_text = (! (always_text | out_quiet)
+	      && memchr (bufbeg, '\0', buflim - bufbeg));
+  done_on_match += not_text;
+  out_quiet += not_text;
+
   for (;;)
     {
-      if (fillbuf(save) < 0)
-	{
-	  error(filename, errno);
-	  return nlines;
-	}
       lastnl = bufbeg;
       if (lastout)
 	lastout = bufbeg;
@@ -556,11 +687,11 @@ grep(fd)
       residue = buflim - lim;
       if (beg < lim)
 	{
-	  nlines += grepbuf(beg, lim);
+	  nlines += grepbuf (beg, lim);
 	  if (pending)
-	    prpending(lim);
+	    prpending (lim);
 	  if (nlines && done_on_match && !out_invert)
-	    return nlines;
+	    goto finish_grep;
 	}
       i = 0;
       beg = lim;
@@ -576,17 +707,187 @@ grep(fd)
       save = residue + lim - beg;
       totalcc += buflim - bufbeg - save;
       if (out_line)
-	nlscan(beg);
+	nlscan (beg);
+      if (fillbuf (save, stats) < 0)
+	{
+	  if (! (is_EISDIR (errno, file) && suppress_errors))
+	    error (filename, errno);
+	  goto finish_grep;
+	}
     }
   if (residue)
     {
-      nlines += grepbuf(bufbeg + save - residue, buflim);
+      nlines += grepbuf (bufbeg + save - residue, buflim);
       if (pending)
-	prpending(buflim);
+	prpending (buflim);
     }
+
+ finish_grep:
+  done_on_match -= not_text;
+  out_quiet -= not_text;
+  if ((not_text & ~out_quiet) && nlines != 0)
+    printf (_("Binary file %s matches\n"), filename);
   return nlines;
 }
 
+static int
+grepfile (file, stats)
+     char const *file;
+     struct stats *stats;
+{
+  int desc;
+  int count;
+  int status;
+
+  if (! file)
+    {
+      desc = 0;
+      filename = _("(standard input)");
+    }
+  else
+    {
+      desc = open (file, O_RDONLY);
+
+      if (desc < 0)
+	{
+	  int e = errno;
+	    
+	  if (is_EISDIR (e, file) && directories == RECURSE_DIRECTORIES)
+	    {
+	      if (stat (file, &stats->stat) != 0)
+		{
+		  error (file, errno);
+		  return 1;
+		}
+
+	      return grepdir (file, stats);
+	    }
+	      
+	  if (!suppress_errors)
+	    {
+	      if (directories == SKIP_DIRECTORIES)
+		switch (e)
+		  {
+#ifdef EISDIR
+		  case EISDIR:
+		    return 1;
+#endif
+		  case EACCES:
+		    /* When skipping directories, don't worry about
+		       directories that can't be opened.  */
+		    if (stat (file, &stats->stat) == 0
+			&& S_ISDIR (stats->stat.st_mode))
+		      return 1;
+		    break;
+		  }
+
+	      error (file, e);
+	    }
+
+	  return 1;
+	}
+
+      filename = file;
+    }
+
+#if O_BINARY
+  /* Set input to binary mode.  Pipes are simulated with files
+     on DOS, so this includes the case of "foo | grep bar".  */
+  if (!isatty (desc))
+    SET_BINARY (desc);
+#endif
+
+  count = grep (desc, file, stats);
+  if (count < 0)
+    status = count + 2;
+  else
+    {
+      if (count_matches)
+	{
+	  if (out_file)
+	    printf ("%s:", filename);
+	  printf ("%d\n", count);
+	}
+
+      if (count)
+	{
+	  status = 0;
+	  if (list_files == 1)
+	    printf ("%s\n", filename);
+	}
+      else
+	{
+	  status = 1;
+	  if (list_files == -1)
+	    printf ("%s\n", filename);
+	}
+
+      if (file && close (desc) != 0)
+	error (file, errno);
+    }
+
+  return status;
+}
+
+static int
+grepdir (dir, stats)
+     char const *dir;
+     struct stats *stats;
+{
+  int status = 1;
+  struct stats *ancestor;
+  char *name_space;
+
+  for (ancestor = stats;  (ancestor = ancestor->parent) != 0;  )
+    if (! ((ancestor->stat.st_ino ^ stats->stat.st_ino)
+	   | (ancestor->stat.st_dev ^ stats->stat.st_dev)))
+      {
+	if (!suppress_errors)
+	  fprintf (stderr, _("%s: warning: %s: %s\n"), prog, dir,
+		   _("recursive directory loop"));
+	return 1;
+      }
+
+  name_space = savedir (dir, (unsigned) stats->stat.st_size);
+
+  if (! name_space)
+    {
+      if (errno)
+	{
+	  if (!suppress_errors)
+	    error (dir, errno);
+	}
+      else
+	fatal (_("Memory exhausted"), 0);
+    }
+  else
+    {
+      size_t dirlen = strlen (dir);
+      int needs_slash = ! (dirlen == FILESYSTEM_PREFIX_LEN (dir)
+			   || IS_SLASH (dir[dirlen - 1]));
+      char *file = NULL;
+      char *namep = name_space;
+      struct stats child;
+      child.parent = stats;
+      out_file += !no_filenames;
+      while (*namep)
+	{
+	  size_t namelen = strlen (namep);
+	  file = xrealloc (file, dirlen + 1 + namelen + 1);
+	  strcpy (file, dir);
+	  file[dirlen] = '/';
+	  strcpy (file + dirlen + needs_slash, namep);
+	  namep += namelen + 1;
+	  status &= grepfile (file, &child);
+	}
+      out_file -= !no_filenames;
+      if (file)
+        free (file);
+      free (name_space);
+    }
+
+  return status;
+}
 
 static void
 usage(status)
@@ -627,6 +928,10 @@ Output control:\n\
   -H, --with-filename       print the filename for each match\n\
   -h, --no-filename         suppress the prefixing filename on output\n\
   -q, --quiet, --silent     suppress all normal output\n\
+  -a, --text                do not suppress binary output\n\
+  -d, --directories=ACTION  how to handle directories\n\
+                            ACTION is 'read', 'recurse', or 'skip'.\n\
+  -r, --recursive           equivalent to --directories=recurse.\n\
   -L, --files-without-match only print FILE names containing no match\n\
   -l, --files-with-matches  only print FILE names containing matches\n\
   -c, --count               only print a count of matching lines per FILE\n"));
@@ -635,8 +940,9 @@ Output control:\n\
 Context control:\n\
   -B, --before-context=NUM  print NUM lines of leading context\n\
   -A, --after-context=NUM   print NUM lines of trailing context\n\
-  -NUM                      same as both -B NUM and -A NUM\n\
-  -C, --context             same as -2\n\
+  -C, --context[=NUM]       print NUM (default 2) lines of output context\n\
+                            unless overriden by -A or -B\n\
+  -NUM                      same as --context=NUM\n\
   -U, --binary              do not strip CR characters at EOL (MSDOS)\n\
   -u, --unix-byte-offsets   report offsets as if CRs were not there (MSDOS)\n\
 \n\
@@ -646,14 +952,14 @@ two FILEs given, assume -h. Exit with 0 if matches, with 1 if none.\n\
 Exit with 2 if syntax errors or system errors.\n"));
       printf (_("\nReport bugs to <bug-gnu-utils@gnu.org>.\n"));
     }
-  exit(status);
+  exit (status);
 }
 
 /* Go through the matchers vector and look for the specified matcher.
    If we find it, install it in compile and execute, and return 1.  */
 static int
-setmatcher(name)
-     char *name;
+setmatcher (name)
+     char const *name;
 {
   int i;
 #ifdef HAVE_SETRLIMIT
@@ -661,7 +967,7 @@ setmatcher(name)
 #endif
 
   for (i = 0; matchers[i].name; ++i)
-    if (strcmp(name, matchers[i].name) == 0)
+    if (strcmp (name, matchers[i].name) == 0)
       {
 	compile = matchers[i].compile;
 	execute = matchers[i].execute;
@@ -696,22 +1002,23 @@ setmatcher(name)
 }
 
 int
-main(argc, argv)
+main (argc, argv)
      int argc;
      char *argv[];
 {
   char *keys;
   size_t keycc, oldcc, keyalloc;
-  int count_matches, no_filenames, list_files, suppress_errors;
   int with_filenames;
-  int opt, cc, desc, count, status;
+  int opt, cc, status;
+  unsigned digit_args_val, default_context;
   FILE *fp;
   extern char *optarg;
   extern int optind;
 
+  initialize_main (&argc, &argv);
   prog = argv[0];
-  if (prog && strrchr(prog, '/'))
-    prog = strrchr(prog, '/') + 1;
+  if (prog && strrchr (prog, '/'))
+    prog = strrchr (prog, '/') + 1;
 
 #if defined(__MSDOS__) || defined(_WIN32)
   /* DOS and MS-Windows use backslashes as directory separators, and usually
@@ -719,7 +1026,7 @@ main(argc, argv)
   if (prog)
     {
       char *p = prog;
-      char *bslash = strrchr(argv[0], '\\');
+      char *bslash = strrchr (argv[0], '\\');
 
       if (bslash && bslash >= prog) /* for mixed forward/backslash case */
 	prog = bslash + 1;
@@ -733,20 +1040,24 @@ main(argc, argv)
 	  *p += 'a' - 'A';
 
       /* Remove the .exe extension, if any.  */
-      if ((p = strrchr(prog, '.')) && strcmp(p, ".exe") == 0)
+      if ((p = strrchr (prog, '.')) && strcmp (p, ".exe") == 0)
 	*p = '\0';
     }
 #endif
 
   keys = NULL;
   keycc = 0;
-  count_matches = 0;
-  no_filenames = 0;
   with_filenames = 0;
-  list_files = 0;
-  suppress_errors = 0;
   matcher = NULL;
 
+  /* The value -1 means to use DEFAULT_CONTEXT. */
+  out_after = out_before = -1;
+  /* Default before/after context: chaged by -C/-NUM options */
+  default_context = 0;
+  /* Accumulated value of individual digits in a -NUM option */
+  digit_args_val = 0;
+
+
 /* Internationalization. */
 #if HAVE_SETLOCALE
   setlocale (LC_ALL, "");
@@ -756,11 +1067,11 @@ main(argc, argv)
   textdomain (PACKAGE);
 #endif
 
-  while ((opt = getopt_long(argc, argv,
+  while ((opt = getopt_long (argc, argv,
 #if O_BINARY
-         "0123456789A:B:CEFGHVX:bce:f:hiLlnqsvwxyUu",
+         "0123456789A:B:C::EFGHVX:abcd:e:f:hiLlnqrsvwxyUu",
 #else
-         "0123456789A:B:CEFGHVX:bce:f:hiLlnqsvwxy",
+         "0123456789A:B:C::EFGHVX:abcd:e:f:hiLlnqrsvwxy",
 #endif
          long_options, NULL)) != EOF)
     switch (opt)
@@ -775,8 +1086,8 @@ main(argc, argv)
       case '7':
       case '8':
       case '9':
-	out_before = 10 * out_before + opt - '0';
-	out_after = 10 * out_after + opt - '0';
+	digit_args_val = 10 * digit_args_val + opt - '0';
+	default_context = digit_args_val;
 	break;
       case 'A':
 	if (optarg)
@@ -793,11 +1104,19 @@ main(argc, argv)
 	  }
 	break;
       case 'C':
-	out_before = out_after = 2;
+	/* Set output match context, but let any explicit leading or
+	   trailing amount specified with -A or -B stand. */
+	if (optarg)
+	  {
+	    if (ck_atoi (optarg, &default_context))
+	      fatal (_("invalid context length argument"), 0);
+	  }
+	else
+	  default_context = 2;
 	break;
       case 'E':
-	if (matcher && strcmp(matcher, "posix-egrep") != 0)
-	  fatal(_("you may specify only one of -E, -F, or -G"), 0);
+	if (matcher && strcmp (matcher, "posix-egrep") != 0)
+	  fatal (_("you may specify only one of -E, -F, or -G"), 0);
 	matcher = "posix-egrep";
 	break;
       case 'F':
@@ -806,8 +1125,8 @@ main(argc, argv)
 	matcher = "fgrep";
 	break;
       case 'G':
-	if (matcher && strcmp(matcher, "grep") != 0)
-	  fatal(_("you may specify only one of -E, -F, or -G"), 0);
+	if (matcher && strcmp (matcher, "grep") != 0)
+	  fatal (_("you may specify only one of -E, -F, or -G"), 0);
 	matcher = "grep";
 	break;
       case 'H':
@@ -826,9 +1145,12 @@ main(argc, argv)
 	break;
       case 'X':
 	if (matcher)
-	  fatal(_("matcher already specified"), 0);
+	  fatal (_("matcher already specified"), 0);
 	matcher = optarg;
 	break;
+      case 'a':
+	always_text = 1;
+	break;
       case 'b':
 	out_byte = 1;
 	break;
@@ -836,27 +1158,37 @@ main(argc, argv)
 	out_quiet = 1;
 	count_matches = 1;
 	break;
+      case 'd':
+	if (strcmp (optarg, "read") == 0)
+	  directories = READ_DIRECTORIES;
+	else if (strcmp (optarg, "skip") == 0)
+	  directories = SKIP_DIRECTORIES;
+	else if (strcmp (optarg, "recurse") == 0)
+	  directories = RECURSE_DIRECTORIES;
+	else
+	  fatal (_("unknown directories method"), 0);
+	break;
       case 'e':
-	cc = strlen(optarg);
-	keys = xrealloc(keys, keycc + cc + 1);
-	strcpy(&keys[keycc], optarg);
+	cc = strlen (optarg);
+	keys = xrealloc (keys, keycc + cc + 1);
+	strcpy (&keys[keycc], optarg);
 	keycc += cc;
 	keys[keycc++] = '\n';
 	break;
       case 'f':
-	fp = strcmp(optarg, "-") != 0 ? fopen(optarg, "r") : stdin;
+	fp = strcmp (optarg, "-") != 0 ? fopen (optarg, "r") : stdin;
 	if (!fp)
-	  fatal(optarg, errno);
+	  fatal (optarg, errno);
 	for (keyalloc = 1; keyalloc <= keycc + 1; keyalloc *= 2)
 	  ;
-	keys = xrealloc(keys, keyalloc);
+	keys = xrealloc (keys, keyalloc);
 	oldcc = keycc;
-	while (!feof(fp)
-	       && (cc = fread(keys + keycc, 1, keyalloc - 1 - keycc, fp)) > 0)
+	while (!feof (fp)
+	       && (cc = fread (keys + keycc, 1, keyalloc - 1 - keycc, fp)) > 0)
 	  {
 	    keycc += cc;
 	    if (keycc == keyalloc - 1)
-	      keys = xrealloc(keys, keyalloc *= 2);
+	      keys = xrealloc (keys, keyalloc *= 2);
 	  }
 	if (fp != stdin)
 	  fclose(fp);
@@ -890,6 +1222,9 @@ main(argc, argv)
 	done_on_match = 1;
 	out_quiet = 1;
 	break;
+      case 'r':
+	directories = RECURSE_DIRECTORIES;
+	break;
       case 's':
 	suppress_errors = 1;
 	break;
@@ -906,16 +1241,21 @@ main(argc, argv)
 	/* long options */
 	break;
       default:
-	usage(2);
+	usage (2);
 	break;
       }
 
+  if (out_after < 0)
+    out_after = default_context;
+  if (out_before < 0)
+    out_before = default_context;
+
   if (show_version)
     {
       printf (_("grep (GNU grep) %s\n"), VERSION);
       printf ("\n");
       printf (_("\
-Copyright (C) 1988, 92, 93, 94, 95, 96, 97 Free Software Foundation, Inc.\n"));
+Copyright (C) 1988, 1992-1998, 1999 Free Software Foundation, Inc.\n"));
       printf (_("\
 This is free software; see the source for copying conditions. There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"));
@@ -924,114 +1264,61 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"))
     }
 
   if (show_help)
-    usage(0);
+    usage (0);
 
   if (keys)
     {
-      /* No keys were specified (e.g. -f /dev/null)/ Match nothing/ */
       if (keycc == 0)
+	/* No keys were specified (e.g. -f /dev/null).  Match nothing.  */
         out_invert ^= 1;
-      else /* strip trailin newline. */
+      else
+	/* Strip trailing newline. */
         --keycc;
     }
   else
     if (optind < argc)
       {
 	keys = argv[optind++];
-	keycc = strlen(keys);
+	keycc = strlen (keys);
       }
     else
-      usage(2);
+      usage (2);
 
-  if (!matcher)
-    matcher = prog;
+  if (! matcher)
+    matcher = default_matcher;
 
-  if (!setmatcher(matcher) && !setmatcher("default"))
-    abort();
+  if (!setmatcher (matcher) && !setmatcher ("default"))
+    abort ();
 
   (*compile)(keys, keycc);
 
   if ((argc - optind > 1 && !no_filenames) || with_filenames)
     out_file = 1;
 
-  status = 1;
+#if O_BINARY
+  /* Output is set to binary mode because we shouldn't convert
+     NL to CR-LF pairs, especially when grepping binary files.  */
+  if (!isatty (1))
+    SET_BINARY (1);
+#endif
+
 
   if (optind < argc)
-    while (optind < argc)
-      {
-	if (strcmp(argv[optind], "-") == 0)
-	  {
-	    filename = _("(standard input)");
-	    desc = 0;
-	  }
-	else
-	  {
-#if defined(__MSDOS__) || defined(_WIN32)
-	    struct stat st;
-	    if (stat (argv[optind], &st) == 0 && S_ISDIR(st.st_mode))
-	      {
-		++optind;
-		continue;
-	      }
-#endif
-	    filename = argv[optind];
-	    desc = open(argv[optind], O_RDONLY);
-	  }
-	if (desc < 0)
-	  {
-	    if (!suppress_errors)
-	      error(argv[optind], errno);
-	  }
-	else
-	  {
-#if O_BINARY
-	    /* Set input to binary mode.  Pipes are simulated with files
-	       on DOS, so this includes the case of "foo | grep bar".  */
-	    if (!isatty(desc))
-	      SET_BINARY(desc);
-#endif
-	    count = grep(desc);
-	    if (count_matches)
-	      {
-		if (out_file)
-		  printf("%s:", filename);
-		printf("%d\n", count);
-	      }
-	    if (count)
-	      {
-		status = 0;
-		if (list_files == 1)
-		  printf("%s\n", filename);
-	      }
-	    else if (list_files == -1)
-	      printf("%s\n", filename);
-	    if (desc != 0)
-	      close(desc);
-	  }
-	++optind;
-      }
-  else
     {
-      filename = _("(standard input)");
-#if O_BINARY
-      if (!isatty(0))
-	SET_BINARY(0);
-#endif
-      count = grep(0);
-      if (count_matches)
-	printf("%d\n", count);
-      if (count)
+	status = 1;
+	do
 	{
-	  status = 0;
-	  if (list_files == 1)
-	    printf(_("(standard input)\n"));
+	  char *file = argv[optind];
+	  status &= grepfile (strcmp (file, "-") == 0 ? (char *) NULL : file,
+			      &stats_base);
 	}
-      else if (list_files == -1)
-	printf(_("(standard input)\n"));
+	while ( ++optind < argc);
     }
+  else
+    status = grepfile ((char *) NULL, &stats_base);
 
   if (fclose (stdout) == EOF)
     error (_("writing output"), errno);
 
-  exit(errseen ? 2 : status);
+  exit (errseen ? 2 : status);
 }
