@@ -41,13 +41,10 @@ static void kwsinit PARAMS((void));
 struct matcher matchers[] = {
   { "default", Gcompile, EGexecute },
   { "grep", Gcompile, EGexecute },
-  { "ggrep", Gcompile, EGexecute },
   { "egrep", Ecompile, EGexecute },
   { "posix-egrep", Ecompile, EGexecute },
   { "awk", Ecompile, EGexecute },
-  { "gegrep", Ecompile, EGexecute },
   { "fgrep", Fcompile, Fexecute },
-  { "gfgrep", Fcompile, Fexecute },
   { 0, 0, 0 },
 };
 
@@ -142,8 +139,6 @@ Gcompile(pattern, size)
   if ((err = re_compile_pattern(pattern, size, &regex)) != 0)
     fatal(err, 0);
 
-  dfainit(&dfa);
-
   /* In the match_words and match_lines cases, we use a different pattern
      for the DFA matcher that will quickly throw out cases that won't work.
      Then if DFA succeeds we do some hairy stuff using the regex matcher
@@ -210,8 +205,6 @@ Ecompile(pattern, size)
   if ((err = re_compile_pattern(pattern, size, &regex)) != 0)
     fatal(err, 0);
 
-  dfainit(&dfa);
-
   /* In the match_words and match_lines cases, we use a different pattern
      for the DFA matcher that will quickly throw out cases that won't work.
      Then if DFA succeeds we do some hairy stuff using the regex matcher
@@ -330,8 +323,9 @@ EGexecute(buf, size, endp)
 	  if (match_words)
 	    while (start >= 0)
 	      {
-		if ((start == 0 || !WCHAR(beg[start - 1]))
-		    && (len == end - beg || !WCHAR(beg[start + len])))
+		if ((start == 0 || !WCHAR ((unsigned char) beg[start - 1]))
+		    && (len == end - beg
+			|| !WCHAR ((unsigned char) beg[start + len])))
 		  goto success;
 		if (len > 0)
 		  {
