@@ -31,7 +31,6 @@
 # define PARAMS(x) ()
 #endif
 
-#define valloc system_valloc /* kludge */
 #ifdef HAVE_UNISTD_H
 # include <fcntl.h>
 # include <unistd.h>
@@ -40,15 +39,25 @@
 int open(), read(), close();
 #endif
 
+#include <errno.h>
+#ifndef errno
+extern int errno;
+#endif
+
+#ifndef HAVE_STRERROR
+extern int sys_nerr;
+extern char *sys_errlist[];
+# define strerror(E) (0 <= (E) && (E) < sys_nerr ? _(sys_errlist[E]) : _("Unknown system error"))
+#endif
+
 /* Some operating systems treat text and binary files differently.  */
-#if defined(O_BINARY) && O_BINARY
+#if O_BINARY
 # include <io.h>
-# ifdef __DJGPP__
-#  define SET_BINARY(fd)  setmode(fd,O_BINARY)
+# ifdef HAVE_SETMODE
+#  define SET_BINARY(fd)  setmode (fd, O_BINARY)
 # else
-#  define SET_BINARY(fd)  _setmode(fd,O_BINARY)
+#  define SET_BINARY(fd)  _setmode (fd, O_BINARY)
 # endif
-static inline int undossify_input PARAMS((char *, size_t));
 #else
 # ifndef O_BINARY
 #  define O_BINARY 0
@@ -56,6 +65,38 @@ static inline int undossify_input PARAMS((char *, size_t));
 # endif
 #endif
 
+#ifdef HAVE_DOS_FILE_NAMES
+# define IS_SLASH(c) ((c) == '/' || (c) == '\\')
+# define FILESYSTEM_PREFIX_LEN(f) ((f)[0] && (f)[1] == ':' ? 2 : 0)
+#endif
+
+#ifndef IS_SLASH
+# define IS_SLASH(c) ((c) == '/')
+#endif
+
+#ifndef FILESYSTEM_PREFIX_LEN
+# define FILESYSTEM_PREFIX_LEN(f) 0
+#endif
+
+/* This assumes _WIN32, like DJGPP, has D_OK.  Does it?  In what header?  */
+#ifdef D_OK
+# ifdef EISDIR
+#  define is_EISDIR(e, f) \
+     ((e) == EISDIR \
+      || ((e) == EACCES && access (f, D_OK) == 0 && ((e) = EISDIR, 1)))
+# else
+#  define is_EISDIR(e, f) ((e) == EACCES && access (f, D_OK) == 0)
+# endif
+#endif
+
+#ifndef is_EISDIR
+# ifdef EISDIR
+#  define is_EISDIR(e, f) ((e) == EISDIR)
+# else
+#  define is_EISDIR(e, f) 0
+# endif
+#endif
+
 #if STAT_MACROS_BROKEN
 # undef S_ISDIR
 #endif
@@ -70,21 +111,15 @@ ptr_t malloc(), realloc(), calloc();
 void free();
 #endif
 
-/* Be careful declaring valloc.  OSF/1 3.2 declares it in stdlib.h,
-   BSD/OS 2.1 in unistd.h.  Ultrix 4.4 doesn't declare it.  */
-#undef valloc /* undo the kludge */
-#if !defined(HAVE_VALLOC)
-# define valloc malloc
-#else
-ptr_t valloc();
-#endif
-
 #if __STDC__
 # include <stddef.h>
 #endif
 #ifdef STDC_HEADERS
 # include <limits.h>
 #endif
+#ifndef CHAR_BIT
+# define CHAR_BIT 8
+#endif
 #ifndef INT_MAX
 # define INT_MAX 2147483647
 #endif
@@ -110,49 +145,30 @@ ptr_t valloc();
 ptr_t memchr();
 #endif
 
-#include <errno.h>
-#ifndef errno
-extern int errno;
-#endif
-
-#ifndef HAVE_STRERROR
-extern int sys_nerr;
-extern char *sys_errlist[];
-# define strerror(E) ((E) < sys_nerr ? sys_errlist[(E)] : "bogus error number")
-#endif
-
 #include <ctype.h>
 
 #ifndef isgraph
 # define isgraph(C) (isprint(C) && !isspace(C))
 #endif
 
-#ifdef isascii
-# define ISALPHA(C) (isascii(C) && isalpha(C))
-# define ISUPPER(C) (isascii(C) && isupper(C))
-# define ISLOWER(C) (isascii(C) && islower(C))
-# define ISDIGIT(C) (isascii(C) && isdigit(C))
-# define ISXDIGIT(C) (isascii(C) && isxdigit(C))
-# define ISSPACE(C) (isascii(C) && isspace(C))
-# define ISPUNCT(C) (isascii(C) && ispunct(C))
-# define ISALNUM(C) (isascii(C) && isalnum(C))
-# define ISPRINT(C) (isascii(C) && isprint(C))
-# define ISGRAPH(C) (isascii(C) && isgraph(C))
-# define ISCNTRL(C) (isascii(C) && iscntrl(C))
+#if defined (STDC_HEADERS) || (!defined (isascii) && !defined (HAVE_ISASCII))
+# define IN_CTYPE_DOMAIN(c) 1
 #else
-# define ISALPHA(C) isalpha(C)
-# define ISUPPER(C) isupper(C)
-# define ISLOWER(C) islower(C)
-# define ISDIGIT(C) isdigit(C)
-# define ISXDIGIT(C) isxdigit(C)
-# define ISSPACE(C) isspace(C)
-# define ISPUNCT(C) ispunct(C)
-# define ISALNUM(C) isalnum(C)
-# define ISPRINT(C) isprint(C)
-# define ISGRAPH(C) isgraph(C)
-# define ISCNTRL(C) iscntrl(C)
+# define IN_CTYPE_DOMAIN(c) isascii(c)
 #endif
 
+#define ISALPHA(C)	(IN_CTYPE_DOMAIN (C) && isalpha (C))
+#define ISUPPER(C)	(IN_CTYPE_DOMAIN (C) && isupper (C))
+#define ISLOWER(C)	(IN_CTYPE_DOMAIN (C) && islower (C))
+#define ISDIGIT(C)	(IN_CTYPE_DOMAIN (C) && isdigit (C))
+#define ISXDIGIT(C)	(IN_CTYPE_DOMAIN (C) && isxdigit (C))
+#define ISSPACE(C)	(IN_CTYPE_DOMAIN (C) && isspace (C))
+#define ISPUNCT(C)	(IN_CTYPE_DOMAIN (C) && ispunct (C))
+#define ISALNUM(C)	(IN_CTYPE_DOMAIN (C) && isalnum (C))
+#define ISPRINT(C)	(IN_CTYPE_DOMAIN (C) && isprint (C))
+#define ISGRAPH(C)	(IN_CTYPE_DOMAIN (C) && isgraph (C))
+#define ISCNTRL(C)	(IN_CTYPE_DOMAIN (C) && iscntrl (C))
+
 #define TOLOWER(C) (ISUPPER(C) ? tolower(C) : (C))
 
 #if ENABLE_NLS
@@ -166,3 +182,7 @@ extern char *sys_errlist[];
 #if HAVE_SETLOCALE
 # include <locale.h>
 #endif
+
+#ifndef initialize_main
+#define initialize_main(argcp, argvp)
+#endif
